import { HashMap, StdAsyncCallback } from "../common";
export interface IMemoryTable<T> {
    /**
     * Add a value if the specific key if not existing.
     *
     * Params:
     *
     * - `EXPIRES: Date|number`
     *
     * If EXPIRES is typeof Date, the new key will be expired
     * at that time.
     *
     * Or if EXPIRES is less then milliseconds of one year,
     * the new key will be expired in EXPIRES ms; if EXPIRES
     * is larger then milliseconds of one year, it will be
     * considered as a Unix timestamp (ms), and the new key
     * will be expired at that time.
     */
    add(key: string, value: T, expires?: any): boolean;
    /**
     * Replace a value of specific key if existing.
     */
    replace(key: string, value: T, expires?: any): boolean;
    /**
     * Set the value for a specific key.
     */
    set(key: string, value: T, expires?: any): IMemoryTable<T>;
    /**
     * Get the time-to-live of a key.
     * If specific key will not expire, return -1.
     * If specific key doesn't exist, return undefined.
     */
    ttl(key: string): number;
    /**
     * Set the new value for time-to-live of a key.
     * If newTTL === null, the key will be keep forever.
     */
    ttl(key: string, newTTL: number): boolean;
    /**
     * Get the expiring timestamp of a key.
     * If specific key will not expire, return -1.
     * If specific key doesn't exist, return undefined.
     */
    /**
     * Get the value of a specific key.
     */
    get(key: string): T;
    /**
     * Get the values of specific keys.
     */
    get(key: string[]): HashMap<T>;
    /**
     * Check if a key exists in this table.
     */
    exist(key: string): boolean;
    /**
     * Check if specific keys exists in this table.
     * Return true only when all specific keys exist.
     */
    exist(key: string[]): boolean;
    /**
     * Delete a key from this table.
     */
    delete(key: string): IMemoryTable<T>;
    /**
     * Delete multi-keys from this table.
     */
    delete(key: string[]): IMemoryTable<T>;
    /**
     * Delete all data from this table.
     */
    truncate(): IMemoryTable<T>;
    /**
     * Number of keys in this table.
     */
    readonly length: number;
    /**
     * Get all keys in this table.
     */
    readonly keys: string[];
    /**
     * Execute the garbage collector.
     */
    gc(asyncCallback?: StdAsyncCallback): void;
}
export interface IMemoryStorage {
    /**
     * Get a typed-table by name.
     */
    get<T>(name: string): IMemoryTable<T>;
    /**
     * Get a table by name.
     */
    get(name: string): IMemoryTable<any>;
    /**
     * Create a table of specific type.
     */
    create<T>(name: string): boolean;
    /**
     * Create a table.
     */
    create(name: string): boolean;
    /**
     * Destroy a table.
     */
    drop(name: string): boolean;
}
/**
 * Memory Stroage Manager
 */
export declare let memory: IMemoryStorage;
