"use strict";
var async = require("async");
var MemoryTable = (function () {
    function MemoryTable() {
        this.spaces = {};
        this.count = 0;
    }
    Object.defineProperty(MemoryTable.prototype, "length", {
        get: function () {
            return this.count;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MemoryTable.prototype, "keys", {
        get: function () {
            var ret = [];
            for (var k in this.spaces) {
                ret.push(k);
            }
            return ret;
        },
        enumerable: true,
        configurable: true
    });
    MemoryTable.prototype.getExpires = function (expires) {
        if (expires instanceof Date) {
            return expires.getTime();
        }
        else if (typeof expires === "number") {
            if (expires <= 31536000000) {
                expires = new Date().getTime() + expires;
            }
            return expires;
        }
        else {
            return undefined;
        }
    };
    MemoryTable.prototype.get = function (key) {
        var node;
        var now = new Date().getTime();
        if (key instanceof Array) {
            var rtn = {};
            for (var _i = 0, key_1 = key; _i < key_1.length; _i++) {
                var x = key_1[_i];
                if (node = this.spaces[x]) {
                    if (node.expires !== undefined && node.expires <= now) {
                        delete this.spaces[x];
                        this.count--;
                        continue;
                    }
                    rtn[x] = node.value;
                }
            }
            return rtn;
        }
        else {
            if (node = this.spaces[key]) {
                if (node.expires !== undefined && node.expires <= now) {
                    delete this.spaces[key];
                    this.count--;
                    return undefined;
                }
                return node.value;
            }
            else {
                return undefined;
            }
        }
    };
    MemoryTable.prototype.ttl = function (key, newTTL) {
        var node = this.spaces[key];
        if (newTTL !== undefined) {
            if (node) {
                node.expires = newTTL + new Date().getTime();
                return true;
            }
            return false;
        }
        else {
            if (node) {
                if (node.expires !== undefined) {
                    if (node.expires <= new Date().getTime()) {
                        delete this.spaces[key];
                        this.count--;
                        return undefined;
                    }
                    else {
                        return node.expires - new Date().getTime();
                    }
                }
                return -1;
            }
            else {
                return undefined;
            }
        }
    };
    MemoryTable.prototype.exist = function (key) {
        var node;
        var now = new Date().getTime();
        if (key instanceof Array) {
            for (var _i = 0, key_2 = key; _i < key_2.length; _i++) {
                var x = key_2[_i];
                if (node = this.spaces[x]) {
                    if (node.expires !== undefined && node.expires <= now) {
                        delete this.spaces[x];
                        this.count--;
                        return false;
                    }
                }
                else {
                    return false;
                }
            }
            return true;
        }
        else {
            if (node = this.spaces[key]) {
                if (node.expires !== undefined && node.expires <= now) {
                    delete this.spaces[key];
                    this.count--;
                    return false;
                }
                return true;
            }
            return false;
        }
    };
    MemoryTable.prototype.add = function (key, value, expires) {
        if (this.spaces[key] !== undefined) {
            return false;
        }
        this.count++;
        this.spaces[key] = {
            "value": value,
            "expires": this.getExpires(expires)
        };
        return true;
    };
    MemoryTable.prototype.replace = function (key, value, expires) {
        if (this.spaces[key] === undefined) {
            return false;
        }
        this.spaces[key] = {
            "value": value,
            "expires": this.getExpires(expires)
        };
        return true;
    };
    MemoryTable.prototype.set = function (key, value, expires) {
        if (!this.spaces[key]) {
            this.count++;
        }
        this.spaces[key] = {
            "value": value,
            "expires": this.getExpires(expires)
        };
        return this;
    };
    MemoryTable.prototype.delete = function (key) {
        if (key instanceof Array) {
            for (var _i = 0, key_3 = key; _i < key_3.length; _i++) {
                var x = key_3[_i];
                delete this.spaces[x];
                this.count--;
            }
        }
        else {
            delete this.spaces[key];
            this.count--;
        }
        return this;
    };
    MemoryTable.prototype.gc = function (asyncCallback) {
        if (asyncCallback) {
            this.gcAsync(asyncCallback);
        }
        else {
            this.gcSync();
        }
    };
    MemoryTable.prototype.gcSync = function () {
        var now = new Date().getTime();
        for (var key in this.spaces) {
            var item = this.spaces[key];
            if (item.expires !== undefined && item.expires < now) {
                delete this.spaces[key];
                this.count--;
            }
        }
    };
    MemoryTable.prototype.gcAsync = function (asyncCallback) {
        var now = new Date().getTime();
        async.forEachOfSeries(this.spaces, function (item, key, next) {
            if (item && item.expires !== undefined && item.expires < now) {
                delete this.spaces[key];
                this.count--;
            }
            setImmediate(next);
        }.bind(this), asyncCallback);
    };
    MemoryTable.prototype.truncate = function () {
        this.spaces = {};
        this.count = 0;
        return this;
    };
    return MemoryTable;
}());
var MemoryStorage = (function () {
    function MemoryStorage() {
        this.tables = {};
    }
    MemoryStorage.prototype.get = function (name) {
        return this.tables[name];
    };
    MemoryStorage.prototype.drop = function (name) {
        if (this.tables[name]) {
            delete this.tables[name];
            return true;
        }
        return false;
    };
    MemoryStorage.prototype.create = function (name) {
        if (this.tables[name]) {
            return false;
        }
        this.tables[name] = new MemoryTable();
        return true;
    };
    return MemoryStorage;
}());
/**
 * Memory Stroage Manager
 */
exports.memory = new MemoryStorage();
//# sourceMappingURL=memstore.js.map