/// <reference types="node" />
/// <reference types="es6-shim" />
import events = require("../events");
import NodeHTTP = require("http");
import { HashMap, StdAsyncCallback } from "../common";
export declare enum ServerStatus {
    "IDLE" = 0,
    "STARTING" = 1,
    "WORKING" = 2,
    "CLOSING" = 3,
    "CLOSED" = 4,
    "ERROR" = 5,
}
export declare enum HttpStatusCode {
    "CONTINUE" = 100,
    "SWITCHING_PROTOCOLS" = 101,
    "PROCESSING" = 102,
    "OK" = 200,
    "CREATED" = 201,
    "ACCEPTED" = 202,
    "NON_AUTHORITATIVE_INFORMATION" = 203,
    "NO_CONTENT" = 204,
    "RESET_CONTENT" = 205,
    "PARTIAL_CONTENT" = 206,
    "MULTIPLE_CHOICES" = 300,
    "MOVED_PERMANENTLY" = 301,
    "MOVE_TEMPORARILY" = 302,
    "SEE_OTHER" = 303,
    "NOT_MODIFIED" = 304,
    "USE_PROXY" = 305,
    "SWITCH_PROXY" = 306,
    "TEMPORARY_REDIRECT" = 307,
    "BAD_REQUEST" = 400,
    "UNAUTHORIZED" = 401,
    "FORBIDDEN" = 403,
    "NOT_FOUND" = 404,
    "METHOD_NOT_ALLOWED" = 405,
    "NOT_ACCEPTABLE" = 406,
    "PROXY_AUTHENTICATION_REQUIRED" = 407,
    "REQUEST_TIMEOUT" = 408,
    "CONFLICT" = 409,
    "GONE" = 410,
    "LENGTH_REQUIRED" = 411,
    "PRECONDITION_FAILED" = 412,
    "REQUEST_ENTITY_TOO_LARGE" = 413,
    "REQUEST_URI_TOO_LONG" = 414,
    "UNSUPPORTED_MEDIA_TYPE" = 415,
    "REQUESTED_RANGE_NOT_SATISFIABLE" = 416,
    "EXPECTATION_FAILED" = 417,
    "UNPROCESSABLE_ENTITY" = 422,
    "LOCKED" = 423,
    "FAILED_DEPENDENCY" = 424,
    "UNORDERED_COLLECTION" = 425,
    "UPGRADE_REQUIRED" = 426,
    "RETRY_WITH" = 449,
    "INTERNAL_SERVER_ERROR" = 500,
    "NOT_IMPLEMENTED" = 501,
    "BAD_GATEWAY" = 502,
    "SERVICE_UNAVAILABLE" = 503,
    "GATEWAY_TIMEOUT" = 504,
    "HTTP_VERSION_NOT_SUPPORTED" = 505,
    "VARIANT_ALSO_NEGOTIATES" = 506,
    "INSUFFICIENT_STORAGE" = 507,
    "BANDWIDTH_LIMIT_EXCEEDED" = 509,
    "NOT_EXTENDED" = 510,
    "UNPARSEABLE_RESPONSE_HEADERS" = 600,
}
/**
 * The controlling object of a HTTP request.
 */
export interface ServerRequest extends NodeHTTP.IncomingMessage {
    /**
     * The raw data of query-string
     */
    queryString: string;
    /**
     * The decoded data of query-string
     */
    query: HashMap<any>;
    /**
     * The parsed path of request, work for router.
     */
    path: string;
    /**
     * The params of request, from routers and middlewares.
     */
    params: HashMap<any>;
    /**
     * The original path in URI.
     */
    rawPath: string;
    /**
     * The matched router handler.
     */
    handler: RouterHandler;
}
/**
 * The signature of a HTTP request handler.
 */
export interface RouterHandler {
    (request: ServerRequest, response: ServerResponse, server: Server, next: StdAsyncCallback): void;
}
/**
 * The signature of a HTTP request handling middleware.
 */
export interface Middleware {
    (request: ServerRequest, response: ServerResponse, next: events.AsyncEventChainCallback): void;
}
/**
 * Reserved structure definition of the controlling object of a HTTP request.
 */
export interface ServerResponse extends NodeHTTP.ServerResponse {
    /**
     * Specific whether `end` method is called.
     */
    finished: boolean;
}
/**
 * Method for HTTP request if SSL enabled.
 */
export declare enum HTTPStatusWithSSL {
    ENABLED = 0,
    DISABLED = 1,
    REDIRECT = 2,
}
/**
 * The basic Options for a server.
 */
export interface ServerOptions {
    "host"?: string;
    "port"?: number;
    "backlog"?: number;
    /**
     * Enable SSL by Node.js.
     *
     * If HTTPS proxy server, such as Nginx, is used,
     * don't enable this option again, just use the proxy!
     */
    "ssl"?: {
        /**
         * SSL certificate.
         */
        "certicate": string;
        /**
         * Private key for SSL certificate.
         */
        "privateKey": string;
        /**
         * Does HTTP still workable without SSL.
         */
        "http"?: HTTPStatusWithSSL;
    };
}
export declare type HTTPMethodName = "GET" | "POST" | "PATCH" | "PUT" | "DELETE" | "HEAD" | "OPTIONS" | "ALL";
/**
 * This class provides the properties for server.
 */
export declare abstract class ServerProperties extends events.AsyncEventEmitter {
    protected server: NodeHTTP.Server;
    private inHost;
    /**
     * The host of server.
     */
    host: string;
    private inPort;
    /**
     * The port of server.
     */
    port: number;
    private inBacklog;
    /**
     * The backlog of server.
     */
    backlog: number;
    /**
     * The status of server.
     */
    private inStatus;
    /**
     * The status of server.
     */
    readonly status: ServerStatus;
    constructor(options?: ServerOptions);
}
export interface RouterRegisterMethod {
    (rule: string | RegExp, handler: RouterHandler): Server;
}
export declare class Server extends ServerProperties {
    /**
     * Router-table for string-matched handlers.
     */
    private stringRouter;
    /**
     * Router-table for advance-rule-matched handlers.
     */
    private advanceRouter;
    /**
     * Router-table for string-matched handlers.
     */
    private middlewares;
    /**
     * Register a handler for GET method.
     */
    get(rule: string | RegExp, handler: RouterHandler): Server;
    /**
     * Register a handler for POST method.
     */
    post(rule: string | RegExp, handler: RouterHandler): Server;
    /**
     * Register a handler for PUT method.
     */
    put(rule: string | RegExp, handler: RouterHandler): Server;
    /**
     * Register a handler for PATCH method.
     */
    patch(rule: string | RegExp, handler: RouterHandler): Server;
    /**
     * Register a handler for DELETE method.
     */
    delete(rule: string | RegExp, handler: RouterHandler): Server;
    /**
     * Register a handler for HEAD method.
     */
    head(rule: string | RegExp, handler: RouterHandler): Server;
    /**
     * Register a handler for OPTIONS method.
     */
    options(rule: string | RegExp, handler: RouterHandler): Server;
    /**
     * Register a handler for all methods if no others matched.
     */
    all(rule: string | RegExp, handler: RouterHandler): Server;
    /**
     * Register handler for Smart-Route rule.
     */
    private registerSmartRouter(method, exp, handler);
    /**
     * Register handler for String rule.
     */
    private registerStringRouter(method, path, handler);
    /**
     * Register handler for RegExp rule.
     */
    private registerRegExpRouter(method, exp, handler);
    /**
     * Setup a middleware for server.
     */
    use(...args: any[]): Server;
    static create(options?: ServerOptions): Server;
    private constructor(options?);
    /**
     * Start the server.
     */
    start(cb?: StdAsyncCallback): void;
    protected setupRequest(req: ServerRequest): void;
    protected gateway(req: ServerRequest, resp: ServerResponse): void;
    route(req: ServerRequest, resp: ServerResponse): void;
}
