
import events = require("../events");
import NodeURL = require("url");
import NodeHTTP = require("http");
import async = require("async");

import { HashMap, Exception, EmptyCallback, StdAsyncCallback } from "../common";
import SmartRouteRule = require("./SmartRouteRule");

export enum ServerStatus {
    "IDLE", "STARTING", "WORKING", "CLOSING", "CLOSED", "ERROR"
}

export enum HttpStatusCode {
    "CONTINUE" = 100,
    "SWITCHING_PROTOCOLS" = 101,
    "PROCESSING" = 102,
    "OK" = 200,
    "CREATED" = 201,
    "ACCEPTED" = 202,
    "NON_AUTHORITATIVE_INFORMATION" = 203,
    "NO_CONTENT" = 204,
    "RESET_CONTENT" = 205,
    "PARTIAL_CONTENT" = 206,
    "MULTIPLE_CHOICES" = 300,
    "MOVED_PERMANENTLY" = 301,
    "MOVE_TEMPORARILY" = 302,
    "SEE_OTHER" = 303,
    "NOT_MODIFIED" = 304,
    "USE_PROXY" = 305,
    "SWITCH_PROXY" = 306,
    "TEMPORARY_REDIRECT" = 307,
    "BAD_REQUEST" = 400,
    "UNAUTHORIZED" = 401,
    "FORBIDDEN" = 403,
    "NOT_FOUND" = 404,
    "METHOD_NOT_ALLOWED" = 405,
    "NOT_ACCEPTABLE" = 406,
    "PROXY_AUTHENTICATION_REQUIRED" = 407,
    "REQUEST_TIMEOUT" = 408,
    "CONFLICT" = 409,
    "GONE" = 410,
    "LENGTH_REQUIRED" = 411,
    "PRECONDITION_FAILED" = 412,
    "REQUEST_ENTITY_TOO_LARGE" = 413,
    "REQUEST_URI_TOO_LONG" = 414,
    "UNSUPPORTED_MEDIA_TYPE" = 415,
    "REQUESTED_RANGE_NOT_SATISFIABLE" = 416,
    "EXPECTATION_FAILED" = 417,
    "UNPROCESSABLE_ENTITY" = 422,
    "LOCKED" = 423,
    "FAILED_DEPENDENCY" = 424,
    "UNORDERED_COLLECTION" = 425,
    "UPGRADE_REQUIRED" = 426,
    "RETRY_WITH" = 449,
    "INTERNAL_SERVER_ERROR" = 500,
    "NOT_IMPLEMENTED" = 501,
    "BAD_GATEWAY" = 502,
    "SERVICE_UNAVAILABLE" = 503,
    "GATEWAY_TIMEOUT" = 504,
    "HTTP_VERSION_NOT_SUPPORTED" = 505,
    "VARIANT_ALSO_NEGOTIATES" = 506,
    "INSUFFICIENT_STORAGE" = 507,
    "BANDWIDTH_LIMIT_EXCEEDED" = 509,
    "NOT_EXTENDED" = 510,
    "UNPARSEABLE_RESPONSE_HEADERS" = 600
}

/**
 * The controlling object of a HTTP request.
 */
export interface ServerRequest extends NodeHTTP.IncomingMessage {

    /**
     * The raw data of query-string 
     */
    queryString: string;

    /**
     * The decoded data of query-string 
     */
    query: HashMap<any>;

    /**
     * The parsed path of request, work for router.
     */
    path: string;

    /**
     * The params of request, from routers and middlewares.
     */
    params: HashMap<any>;

    /**
     * The original path in URI.
     */
    rawPath: string;

    /**
     * The matched router handler.
     */
    handler: RouterHandler;
}

/**
 * The signature of a HTTP request handler.
 */
export interface RouterHandler {

    (request: ServerRequest, response: ServerResponse, server: Server, next: StdAsyncCallback): void;
}

/**
 * The structure of routing-rule descriptors.
 */
interface AdvanceRouterRule {

    "regExp"?: RegExp;

    "smartExp"?: SmartRouteRule;

    "handler": RouterHandler;
}

/**
 * The structure of middleware descriptors.
 */
interface MiddlewareRule {

    "expr"?: string | RegExp | SmartRouteRule;

    "handler": Middleware;
}

/**
 * The signature of a HTTP request handling middleware.
 */
export interface Middleware {

    (request: ServerRequest, response: ServerResponse, next: events.AsyncEventChainCallback): void;
}

const DEFAULT_HOST: string     = "127.0.0.1";
const DEFAULT_PORT: number     = 80;
const DEFAULT_BACKLOG: number  = 511;

/**
 * Reserved structure definition of the controlling object of a HTTP request.
 */
export interface ServerResponse extends NodeHTTP.ServerResponse {

    /**
     * Specific whether `end` method is called.
     */
    finished: boolean;
}

/**
 * Method for HTTP request if SSL enabled.
 */
export enum HTTPStatusWithSSL {
    ENABLED,
    DISABLED,
    REDIRECT
};

/**
 * The basic Options for a server.
 */
export interface ServerOptions {

    "host"?: string;

    "port"?: number;

    "backlog"?: number;

    /**
     * Enable SSL by Node.js.
     * 
     * If HTTPS proxy server, such as Nginx, is used,
     * don't enable this option again, just use the proxy!
     */
    "ssl"?: {

        /**
         * SSL certificate.
         */
        "certicate": string;

        /**
         * Private key for SSL certificate.
         */
        "privateKey": string;

        /**
         * Does HTTP still workable without SSL.
         */
        "http"?: HTTPStatusWithSSL;
    };
}

/**
 * List of all supported methods.
 */
const AVAILABLE_METHODS: string[] = [
    "GET", "POST", "PATCH", "PUT",
    "DELETE", "HEAD", "OPTIONS", "ALL"
];

export type HTTPMethodName =
    "GET" | "POST" | "PATCH" | "PUT" |
    "DELETE" | "HEAD" | "OPTIONS" | "ALL"; // "ALL" is a trick.

/**
 * Defines the router-table for each method supported.
 */
class RouterTable<T> {

    public "ALL": T;
    public "GET": T;
    public "POST": T;
    public "PUT": T;
    public "PATCH": T;
    public "DELETE": T;
    public "HEAD": T;
    public "OPTIONS": T;
}

/**
 * This class provides the properties for server.
 */
export abstract class ServerProperties extends events.AsyncEventEmitter {

    protected server: NodeHTTP.Server;

    private inHost: string;

    /**
     * The host of server.
     */
    public get host(): string {

        return this.inHost;
    }

    public set host(v: string) {

        if (this.inStatus === ServerStatus.IDLE) {

            this.inHost = v;
        }
    }

    private inPort: number;

    /**
     * The port of server.
     */
    public get port(): number {

        return this.inPort;
    }

    public set port(v: number) {

        if (this.inStatus === ServerStatus.IDLE) {

            this.inPort = v;
        }
    }

    private inBacklog: number;

    /**
     * The backlog of server.
     */
    public get backlog(): number {

        return this.inBacklog;
    }

    public set backlog(v: number) {

        if (this.inStatus === ServerStatus.IDLE) {

            this.inBacklog = v;
        }
    }

    /**
     * The status of server.
     */
    private inStatus: ServerStatus;

    /**
     * The status of server.
     */
    public get status(): ServerStatus {

        return this.inStatus;
    }

    public constructor(options?: ServerOptions) {

        super();

        this.inStatus = ServerStatus.IDLE;

        if (!options) {

            options = {};
        }

        this.inBacklog = options.backlog !== undefined ? options.backlog : DEFAULT_BACKLOG;

        this.inHost = options.host !== undefined ? options.host : DEFAULT_HOST;

        this.inPort = options.port !== undefined ? options.port : DEFAULT_PORT;

    }
}

export interface RouterRegisterMethod {

    (rule: string | RegExp, handler: RouterHandler): Server;
}

export class Server extends ServerProperties {

    // --------- Router Begin ---------

    /**
     * Router-table for string-matched handlers.
     */
    private stringRouter: RouterTable<HashMap<RouterHandler>>;

    /**
     * Router-table for advance-rule-matched handlers.
     */
    private advanceRouter: RouterTable<AdvanceRouterRule[]>;

    /**
     * Router-table for string-matched handlers.
     */
    private middlewares: MiddlewareRule[];

    /**
     * Register a handler for GET method.
     */
    public get(rule: string | RegExp, handler: RouterHandler): Server {

        if (typeof rule === "string") {

            this.registerStringRouter("GET", rule, handler);

        } else if (rule instanceof RegExp) {

            this.registerRegExpRouter("GET", rule, handler);

        }

        return this;
    }

    /**
     * Register a handler for POST method.
     */
    public post(rule: string | RegExp, handler: RouterHandler): Server {

        if (typeof rule === "string") {

            this.registerStringRouter("POST", rule, handler);

        } else if (rule instanceof RegExp) {

            this.registerRegExpRouter("POST", rule, handler);

        }

        return this;
    }

    /**
     * Register a handler for PUT method.
     */
    public put(rule: string | RegExp, handler: RouterHandler): Server {

        if (typeof rule === "string") {

            this.registerStringRouter("PUT", rule, handler);

        } else if (rule instanceof RegExp) {

            this.registerRegExpRouter("PUT", rule, handler);

        }

        return this;
    }

    /**
     * Register a handler for PATCH method.
     */
    public patch(rule: string | RegExp, handler: RouterHandler): Server {

        if (typeof rule === "string") {

            this.registerStringRouter("PATCH", rule, handler);

        } else if (rule instanceof RegExp) {

            this.registerRegExpRouter("PATCH", rule, handler);

        }

        return this;
    }

    /**
     * Register a handler for DELETE method.
     */
    public delete(rule: string | RegExp, handler: RouterHandler): Server {

        if (typeof rule === "string") {

            this.registerStringRouter("DELETE", rule, handler);

        } else if (rule instanceof RegExp) {

            this.registerRegExpRouter("DELETE", rule, handler);

        }

        return this;
    }

    /**
     * Register a handler for HEAD method.
     */
    public head(rule: string | RegExp, handler: RouterHandler): Server {

        if (typeof rule === "string") {

            this.registerStringRouter("HEAD", rule, handler);

        } else if (rule instanceof RegExp) {

            this.registerRegExpRouter("HEAD", rule, handler);

        }

        return this;
    }

    /**
     * Register a handler for OPTIONS method.
     */
    public options(rule: string | RegExp, handler: RouterHandler): Server {

        if (typeof rule === "string") {

            this.registerStringRouter("OPTIONS", rule, handler);

        } else if (rule instanceof RegExp) {

            this.registerRegExpRouter("OPTIONS", rule, handler);

        }

        return this;
    }

    /**
     * Register a handler for all methods if no others matched.
     */
    public all(rule: string | RegExp, handler: RouterHandler): Server {

        if (typeof rule === "string") {

            this.registerStringRouter("ALL", rule, handler);

        } else if (rule instanceof RegExp) {

            this.registerRegExpRouter("ALL", rule, handler);

        }

        return this;
    }

    /**
     * Register handler for Smart-Route rule.
     */
    private registerSmartRouter(method: HTTPMethodName, exp: SmartRouteRule, handler: RouterHandler): void {

        this.advanceRouter[method].push({
            "smartExp": exp,
            "handler": handler
        });
    }

    /**
     * Register handler for String rule.
     */
    private registerStringRouter(method: HTTPMethodName, path: string, handler: RouterHandler): void {

        if (path.length > 1 && path[path.length - 1] === "/") {

            path = path.substr(0, path.length - 1);
        }

        if (path.indexOf("{") > -1) {

            this.registerSmartRouter(method, new SmartRouteRule(path), handler);

        } else {

            this.stringRouter[method][path] = handler;
        }

    }

    /**
     * Register handler for RegExp rule.
     */
    private registerRegExpRouter(method: HTTPMethodName, exp: RegExp, handler: RouterHandler): void {

        this.advanceRouter[method].push({
            "regExp": exp,
            "handler": handler
        });
    }

    // --------- Router End ---------

    // --------- Middleware Begin ---------

    /**
     * Setup a middleware for server.
     */
    public use(...args): Server {

        if (args.length === 0) {

            throw new Exception("BAD-MIDDLEWARE", "Invalid middleware registration.");
        }

        switch (args.length) {

        case 0:

            throw new Exception("BAD-MIDDLEWARE", "Invalid middleware registration.");

        case 1:

            if (args[0] instanceof Function) {

                this.middlewares.push({
                    "handler": args[0]
                });

            } else {

                throw new Exception("BAD-MIDDLEWARE", "Invalid middleware registration.");
            }

            break;

        case 2:
        default:

            if ((args[0] instanceof RegExp || typeof(args[0]) === "string") && args[1] instanceof Function) {

                if (typeof(args[0]) === "string") {

                    if (args[0].indexOf("{") > 0) {

                        args[0] = new SmartRouteRule(args[0]);
                    }
                }

                this.middlewares.push({
                    "expr": args[0],
                    "handler": args[1]
                });

            } else {

                throw new Exception("BAD-MIDDLEWARE", "Invalid middleware registration.");
            }

            break;
        }

        return this;
    }

    // --------- Middleware End ---------

    public static create(options?: ServerOptions): Server {

        let server: any = function(req: ServerRequest, resp: ServerResponse): void {

            (<Server>server).gateway(req, resp);

        };

        Server.apply(server);

        Object.setPrototypeOf(server, Server.prototype);

        return server;
    }

    private constructor(options?: ServerOptions) {

        super(options);

        // Initialize the router table.

        this.stringRouter = new RouterTable<HashMap<RouterHandler>>();

        this.advanceRouter = new RouterTable<AdvanceRouterRule[]>();

        for (let key of AVAILABLE_METHODS) {

            this.stringRouter[key] = {};
            this.advanceRouter[key] = [];
        }

        this.middlewares = [];
    }

    /**
     * Start the server.
     */
    public start(cb?: StdAsyncCallback) {

        let _this: Server = this;

        this.server = NodeHTTP.createServer(this.gateway.bind(this));

        this.server.once("error", function(err) {

            _this.emit("error", [
                new Exception("START-LISTEN", "Failed to listen to server")
            ], cb);
        });

        this.server.listen(this.port, this.host, this.backlog, cb);
    }

    protected setupRequest(req: ServerRequest): void {

        let url: NodeURL.Url = NodeURL.parse(req.url);

        req.rawPath = url.pathname;

        req.queryString = url.query;

        req.params = {};

        if (url.pathname.length === 1) {

            req.path = "/";

        } else {

            if (req.rawPath[req.rawPath.length - 1] === "/") {

                req.path = req.rawPath.substr(0, req.rawPath.length - 1);

            } else {

                req.path = req.rawPath;
            }

            if (req.path[0] !== "/") {

                req.path = "/" + req.path;
            }

        }

    }

    protected gateway(req: ServerRequest, resp: ServerResponse): void {

        this.setupRequest(req);

        if (!this.stringRouter[req.method]) {

            this.emit("/http/405", [req, resp, this], function(err?: Exception) {

                resp.finished || resp.end();
            });

            return;
        }

        req.handler = this.stringRouter[req.method][req.path];

        if (!req.handler) {

            for (let x of this.advanceRouter[req.method]) {

                if (x.regExp) {

                    if (req.path.match(x.regExp)) {

                        req.handler = x.handler;
                        break;
                    }

                } else {

                    if (x.smartExp.match(req.path, req.params)) {

                        req.handler = x.handler;
                        break;
                    }

                }

            }

        }

        if (this.middlewares) {

            async.forEachOfSeries(this.middlewares, function(item: MiddlewareRule, key: number, next: ErrorCallback): void {

                if (item.expr) {

                    if (typeof(item.expr) === "string") {

                        if (item.expr !== req.path) {

                            next();
                            return;
                        }

                    } else if (item.expr instanceof SmartRouteRule) {

                        if (!item.expr.match(req.path, req.params)) {

                            next();
                            return;
                        }

                    } else if (item.expr instanceof RegExp) {

                        if (!req.path.match(item.expr)) {

                            next();
                            return;
                        }

                    }
                }

                item.handler(req, resp, function(err?: Exception, breaked?: boolean): void {

                    if (breaked) {

                        next(new Exception("--break", "Middleware breaked the handling chain."));

                    } else {

                        next(err);
                    }

                });

            }, function(this: Server, err?: Exception) {

                if (err) {

                    if (err.name === "--break") {

                        resp.finished || resp.end();
                        return;

                    } else {

                        return this.emit(err.name, [req, resp, this], function() {

                            resp.finished || resp.end();
                        });
                    }
                }

                this.route(req, resp);

            }.bind(this));

        } else {

            this.route(req, resp);

        }

    }

    public route(req: ServerRequest, resp: ServerResponse): void {

        if (req.handler) {

            let _this: Server = this;

            req.handler(req, resp, this, function(err?: Exception) {

                if (err) {

                    _this.emit(err.name, [req, resp, this], function(err?: Exception) {

                        resp.finished || resp.end();
                    });

                } else {

                    resp.finished || resp.end();
                }

            });

        } else {

            this.emit("/http/404", [req, resp, this], function(err?: Exception) {

                resp.finished || resp.end();
            });

        }

    }

}
