
import { Server, ServerResponse, ServerRequest, Middleware } from "../server";
import NodePath = require("path");
import NodeFS = require("fs");
import async = require("async");

import * as cache from "../../cache";

import events = require("../../events");

import { HashMap, Exception } from "../../common";

interface CacheNode {
    mine: MINEConfig;
    data: Buffer;
}

export interface MINEConfig {
    "MINE": string;
    "cachable"?: boolean;
    "maxCacheSize"?: number;
}

const DEFAULT_BUFFER_SIZE: number = 8096;

export let staticMineTypes: HashMap<MINEConfig> = {
    ".png": {
        "MINE": "image/png"
    },
    ".gif": {
        "MINE": "image/gif"
    },
    ".jpg": {
        "MINE": "image/jpeg"
    },
    ".jpeg": {
        "MINE": "image/jpeg"
    },
    ".bmp": {
        "MINE": "image/bmp"
    },
    ".svg": {
        "MINE": "image/svg+xml"
    },
    ".ico": {
        "MINE": "image/x-icon"
    },
    ".tar": {
        "MINE": "application/x-tar"
    },
    ".tgz": {
        "MINE": "application/x-compressed"
    },
    ".txt": {
        "MINE": "text/plain"
    },
    ".json": {
        "MINE": "application/json",
        "cachable": true
    },
    ".zip": {
        "MINE": "application/zip"
    },
    ".html": {
        "MINE": "text/html",
        "cachable": true
    },
    ".htm": {
        "MINE": "text/html",
        "cachable": true
    },
    ".css": {
        "MINE": "text/css",
        "cachable": true
    },
    ".js": {
        "MINE": "application/x-javascript",
        "cachable": true
    },
    "default": {
        "MINE": "application/octet-stream"
    }
};

export interface StaticDelegateOptions {

    /**
     * The prefix of url. 
     * NOTICE: Must be ending with "/".
     */
    prefix: string;

    /**
     * The root path of static resources.
     * NOTICE: Must be ending with "/".
     */
    root: string;

    /**
     * How long will the static files expire in client. (Seconds)
     */
    expires?: number;

    /**
     * How many bytes of buffer for output.
     */
    bufferSize?: number;

    /**
     * If this option is set, the static files will be cached in
     * memory with the built-in Memory cache engine.
     */
    memoryStorage?: {

        /**
         * Specify an ID for MemoryStorage table.
         */
        table: cache.IMemoryTable<any>;

        /**
         * How many bytes of data could be saved in memory cache.
         */
        maxSize?: number;

        /**
         * How long will the cache items expire, in seconds.
         */
        expires?: number;

        /**
         * How often does the Garbage Collector execute.
         */
        gcInterval?: number;
    };
}

export function StaticDelegate(options: StaticDelegateOptions): Middleware {

    let memStore: cache.IMemoryTable<Buffer>;

    if (options.prefix[options.prefix.length - 1] !== "/") {

        options.prefix += "/";
    }

    if (options.root[options.root.length - 1] !== "/") {

        options.root += "/";
    }

    let expires: number;

    if (options.expires && options.expires > 0) {

        expires = Math.ceil(1000 * options.expires);
    }

    return function(req: ServerRequest, resp: ServerResponse, next: events.AsyncEventChainCallback): void {

        console.log(`Detecting: ${req.url}`);

        if (!req.path.startsWith(options.prefix) || req.path.length === options.prefix.length) {

            return next();
        }

        if (NodePath.normalize(req.path)[0] === ".") {

            return next(new Exception("/http/404", "Suspected URL detected."));
        }

        console.log(`Static Request: ${req.url}`);

        let memStore: cache.IMemoryTable<CacheNode> = options.memoryStorage.table;

        if (memStore) {

            let cn: CacheNode = memStore.get(req.path);

            if (cn) {

                console.log(`Static Cache Hit.`);

                resp.setHeader("Content-Type", cn.mine.MINE);
                resp.setHeader("Content-Length", cn.data.length.toString());

                if (expires) {
                    resp.setHeader("Expires", new Date(Date.now() + expires).toUTCString());
                    resp.setHeader("Cache-Control", `public, max-age=${options.expires}`);
                }

                resp.end(cn.data);

                return next(null, true);
            }
        }

        let fn = options.root + req.path.substr(options.prefix.length);

        let mine: MINEConfig;

        mine = staticMineTypes[NodePath.extname(fn)] || staticMineTypes["default"];

        NodePath.normalize(options.root + req.path.substr(options.prefix.length));

        console.log(`Static Loading "${fn}".`);

        NodeFS.readFile(fn, function(err: Error, data: Buffer): void {

            if (err) {

                return next(new Exception("/http/404", "File not found."));
            }

            resp.setHeader("Content-Type", mine.MINE);
            resp.setHeader("Content-Length", data.length.toString());

            if (expires) {
                resp.setHeader("Expires", new Date(Date.now() + expires).toUTCString());
                resp.setHeader("Cache-Control", `public, max-age=${options.expires}`);
            }

            resp.end(data);

            if (memStore && mine.cachable) {

                memStore.set(req.url, {
                    "mine": mine,
                    "data": data
                });

                console.log(`"${req.path}"" cached.`);
            }

            next(null, true);
        });

    };
}
