import NodeNet = require("net");
import "langext";

let client: NodeNet.Socket = new NodeNet.Socket();
/**
 *   Byte/     0       |       1       |       2       |       3       |
 *      /              |               |               |               |
 *     |0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|
 *     +---------------+---------------+---------------+---------------+
 *    0| Magic         | Opcode        | Key length                    |
 *     +---------------+---------------+---------------+---------------+
 *    4| Extras length | Data type     | vbucket id                    |
 *     +---------------+---------------+---------------+---------------+
 *    8| Total body length                                             |
 *     +---------------+---------------+---------------+---------------+
 *   12| Opaque                                                        |
 *     +---------------+---------------+---------------+---------------+
 *   16| CAS                                                           |
 *     |                                                               |
 *     +---------------+---------------+---------------+---------------+
 *     Total 24 bytes
 */

enum MagicCode {
    REQUEST     = 0x80,
    RESPONSE    = 0x81
}

enum OpCodes {
    GET         = 0,
    SET         = 1,
    ADD         = 2,
    REPLACE     = 3,
    DELETE      = 4,
    INCREMENT   = 5,
    DECREMENT   = 6
}

const MC_HEADER_LENGTH: number = 24;

function createRequestHeader(
    opcode: OpCodes,
    key: string,
    valueLength: number,
    vBucket: number = 0,
    extraLength: number = 0
): Buffer {

    let ret: Buffer = new Buffer(24);
    let keyBuf: Buffer = new Buffer(key);

    ret.writeUInt8(MagicCode.REQUEST, 0);
    ret.writeUInt8(opcode, 1);
    ret.writeUInt16BE(keyBuf.byteLength, 2);
    ret.writeUInt8(extraLength, 4);
    ret.writeUInt8(0, 5);
    ret.writeUInt16BE(vBucket, 6);
    ret.writeUInt32BE(keyBuf.byteLength + extraLength + valueLength, 8);
    ret.writeUInt32BE(123321, 12);
    ret.writeUInt32BE(0, 16);
    ret.writeUInt32BE(0, 20);

    return ret;
}

enum StatusCode {

    NO_ERROR            = 0x0000, // No error
    NOT_FOUND           = 0x0001, // Key not found
    DUP_KEY             = 0x0002, // Key exists
    DATE_TOO_LONG       = 0x0003, // Value too large
    BAD_ARGUMENTS       = 0x0004, // Invalid arguments
    NOT_STORED          = 0x0005, // Item not stored
    NON_NUMERIC         = 0x0006, // Incr/Decr on non-numeric value.
    BAD_VBUCKET         = 0x0007, // The vbucket belongs to another server
    AUTH_ERROR          = 0x0008, // Authentication error
    AUTH_CONTINUE       = 0x0009, // Authentication continue
    UNKNOWN_CMD         = 0x0081, // Unknown command
    OUT_OF_MEMORY       = 0x0082, // Out of memory
    NOT_SUPPORTED       = 0x0083, // Not supported
    INTERNAL_ERROR      = 0x0084, // Internal error
    BUSY                = 0x0085, // Busy
    TEMP_FAILURE        = 0x0086, // Temporary failure
}

/**
 *   Byte/     0       |       1       |       2       |       3       |
 *      /              |               |               |               |
 *     |0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|
 *     +---------------+---------------+---------------+---------------+
 *    0| Magic         | Opcode        | Key Length                    |
 *     +---------------+---------------+---------------+---------------+
 *    4| Extras length | Data type     | Status                        |
 *     +---------------+---------------+---------------+---------------+
 *    8| Total body length                                             |
 *     +---------------+---------------+---------------+---------------+
 *   12| Opaque                                                        |
 *     +---------------+---------------+---------------+---------------+
 *   16| CAS                                                           |
 *     |                                                               |
 *     +---------------+---------------+---------------+---------------+
 *     Total 24 bytes
 */
class ResponsePack {

    public "magic": MagicCode;

    public "opcode": OpCodes;

    public "keyLength": number;

    public "extraLength": number;

    public "dataType": number;

    public "dataLength": number;

    public "status": StatusCode;

    public "opaque": number;

    public "cas": number;

    public "key": string;

    public "extras": Buffer;

    public "data": Buffer;
}

function unpackResponsePack(pack: Buffer): ResponsePack {

    let ret: ResponsePack = new ResponsePack();

    ret.magic = pack.readUInt8(0);

    ret.opcode = pack.readUInt8(1);

    ret.keyLength = pack.readUInt16BE(2);

    ret.extraLength = pack.readUInt8(4);

    ret.dataType = pack.readUInt8(5);

    ret.status = pack.readUInt16BE(6);

    ret.opaque = pack.readUInt32BE(12);

    ret.cas = pack.readUIntBE(16, 8);

    ret.dataLength = pack.readUInt32BE(8) - ret.keyLength - ret.extraLength;

    if (ret.extraLength) {

        ret.extras = pack.slice(
            MC_HEADER_LENGTH,
            MC_HEADER_LENGTH + ret.extraLength
        );
    }

    if (ret.keyLength) {
        let keyData: Buffer = pack.slice(
            MC_HEADER_LENGTH + ret.extraLength,
            MC_HEADER_LENGTH + ret.extraLength + ret.keyLength
        );

        ret.key = keyData.toString();
        keyData = undefined;
    }

    if (ret.dataLength) {

        ret.data = pack.slice(
            MC_HEADER_LENGTH + ret.extraLength + ret.keyLength,
            MC_HEADER_LENGTH + ret.extraLength + ret.keyLength + ret.dataLength
        );

    }

    return ret;
}

const PORT: number = 11211;
const HOST: string = "127.0.0.1";

client.connect(PORT, HOST, function() {

    console.log("CONNECTED TO: " + HOST + ":" + PORT);
    // 建立连接后立即向服务器发送数据，服务器将收到这些数据 
    //*
    client.write(createRequestHeader(OpCodes.GET, "abc", 0));
    client.write(new Buffer("abc"));
    return; //*/
    let extras: Buffer = new Buffer(8);
    extras.writeUInt32BE(0x12345678, 0);
    extras.writeUInt32BE(30, 4);
    client.write(createRequestHeader(OpCodes.SET, "abc", 12, 0, extras.byteLength));
    client.write(extras);
    client.write(new Buffer("abc"));
    client.write(new Buffer("hello world!"));
    // client.write(new Buffer("abc"));

});

client.on("error", function(err) {
    console.log("ERROR", err);
});

// 为客户端添加“data”事件处理函数
// data是服务器发回的数据
client.on("data", function(data: Buffer) {

    let unpackedData: ResponsePack = unpackResponsePack(data);

    console.log(unpackedData);
    unpackedData.data && console.log(unpackedData.data.toString("utf-8"));

    // 完全关闭连接
    client.destroy();

});

// 为客户端添加“close”事件处理函数
client.on("close", function() {
    console.log("Connection closed");
});

